@make(ais)
@aiInfo(topic=<Representation clauses and nonstatic constraints>,
number=[AI-00301/07],
classcode=[ra],
statuscode=[WI],
mainref=[13.02@\(06)],
date=[90-10-19])
@aiHeading()
@begin(history)
@begin(toplines)
@mainref()
@standard[13.04@\(07)]
!class ramification 87-04-17 (provisional classification)
!status work-item 87-01-18 (discussed 89-02-27)
@end(toplines)
!status work-item 87-01-18
!status received 84-10-16
!references AI-00132, AI-00324
@topic{!topic Representation clauses and nonstatic constraints}
@end(history)

@subheading[!summary @cbon()91-02-18@cboff() (DRAFT)]

Any constraint given for an access type does not affect the legality of a
size specification @cbon()for that type or for a composite type having a component
with the specified constraint.  Similarly, such a constraint for a record
component having an access type does not affect the legality of a component
clause for that component.@cboff()

If a component of a record type or @cbon()of an array type does not have an access
type but has a nonstatic constraint,@cboff() no size specification can be given for
the composite type.  Similarly, if a record component @cbon()does not have an access
type but has a nonstatic constraint,@cboff() no component clause can be given for the
component.

If a component of an unconstrained record type has an index or discriminant
constraint that depends on a discriminant, no component clause can be given
for the component, nor can a size specification be given for the record type.

@subheading[!question 88-05-03 (DRAFT)]

13.2(6) states:
@begin(quotation)
The  size  specification  [T'@|SIZE]  is  only   allowed   if   the
constraints on T and on its subcomponents (if any) are static.
@end(quotation)
Similarly, 13.4(7) states (as interpreted by AI-00132):
@begin(quotation)
A component clause is only allowed for a component [if the constraints
(if any) upon this component and its subcomponents are all static].
@end(quotation)
Consider an access type with a nonstatic constraint.  Can a representation
clause be given for a record type with a subcomponent having this access type?
For example:
@begin(example)
@key[package] PK @key[is]
    B : BOOLEAN := TRUE;      -- B is not static
    @key[type] H (D : BOOLEAN) @key[is]
        @key[record]
            @key[null];
        @key[end] @key[record];
@hinge()

    @key[type] T @key[is] @key[access] H (D => B);
    @key[for] T'@|SIZE @key[use] ...;       -- legal? (yes)
@hinge()

    @key[type] U @key[is] @key[access] H;
@hinge()

    @key[type] R @key[is]
        @key[record]
            C : T;
            E : U (B);
        @key[end] @key[record];
    @key[for] R'@|SIZE @key[use] 100;       -- legal? (yes)
@key[end] PK;
@end(example)
Are the size specifications legal?  Is a component clause allowed for
components C and E of record type R?

The discussion in AI-00132 says:
@begin(quotation)
It was intended to allow record representation clauses only for components
whose size was determinable at compile time.
@end(quotation)
If this is indeed the intent, then suppose a record type has a scalar
component with a nonstatic range constraint, e.g.,
@begin(example)
F : INTEGER @key[range] 0 .. BOOLEAN'@|POS(B);
@end(example)
Can a component clause be given for F if the component clause allows enough
space for any INTEGER value?  Suppose the type mark in a component
declaration denoted a static subtype, e.g.,
@begin(example)
@key[type] SMALL @key[is] INTEGER @key[range] 0 .. 7;
@key[type] S @key[is]
    @key[record]
        G : SMALL @key[range] 0 .. BOOLEAN'@|POS(B);
    @key[end] @key[record];
@end(example)
Can a component clause be given for G that allows only SMALL'@|SIZE storage
space?  Can the component clause specify only enough space for an unsigned
representation, since the lower bound is zero and static?

Suppose a component depends on a discriminant:
@begin(example)
@key[type] REC (D : INTEGER) @key[is]
     @key[record]
          C : STRING (1..D);
     @key[end] @key[record];
@hinge()

@key[type] NREC @key[is] @key[new] REC(10);
@end(example)
Does component C of type NREC have a static index constraint?  Can a
component clause be given for component C?  Can a size specification be given
for type NREC?

@subheading[!response @cbon()91-02-18@cboff() (DRAFT)]

The answer to the questions concerning access types depends on the meaning of
a constraint imposed on an access type.  3.3(4) defines @dq"constraint" and
related terms:
@begin(quotation)
The set of possible values for an object of a given type can be
subjected to a condition that is called a @i[constraint] @ellipsis(); a value
is said to @i[satisfy] a constraint if it satisfies the corresponding
condition.  A @i[subtype] is a type together with a constraint; a
value is said to @i[belong to a subtype] of a given type if it belongs
to the type and satisfies the constraint; @ellipsis()
@end(quotation)
For access types, 3.8(6) says:
@begin(quotation)
An access value @i[belongs] to a corresponding subtype of an access
type either if the access value is the null value or if the value of
the designated object satisfies the constraint.
@end(quotation)
AI-00324 noted that this wording was not quite adequate when the designated
type was a scalar type.  AI-00324's recommendation says:
@begin(quotation)
Every access value belongs to its corresponding access type.  An access
value belongs to a subtype of an access type if the access value is the
null value or if the (access) subtype imposes no constraint on the
designated object; if the (access) subtype imposes a constraint (such a
constraint is possible only if the designated type is an array type or a
type that has a discriminant), the access value belongs to the access
subtype if the value of the designated object satisfies the constraint.
@end(quotation)
Given these wordings, a constraint given for an access type, either in the
declaration of the access type (as for type T in the question) or in a
subtype indication (as for component D in the question) applies to the
designated object@dash()the constraint does not restrict the set of access
values except to ensure that when the access value is not @b[null], the
designated object satisfies any index or discriminant constraint given for
the access type.  Therefore, an index or discriminant constraint given for an
access type is not @dq"on" the access type, but instead is @dq"on" the designated
type.  Hence, a size specification is allowed for an access type, even if the
designated type is constrained by a nonstatic constraint, and similarly, any
constraint given for a component that has an access type does not affect the
legality of a component clause for that component or a size clause for the
type to which the component belongs.

This interpretation of an access type constraint is consistent with the
discussion of constraints elsewhere in the standard.  12.3.5(2) says:
@begin(quotation)
If a generic unit has a formal access type, the elaboration of a
corresponding instantiation checks that any @ux[constraints on] the
designated objects are the same for the actual access subtype as for
the formal access type.
@end(quotation)
This wording shows clearly that a constraint given for an access type is
considered to be @dq"on" the designated objects, because it is the values of
these objects that are constrained.  For nonaccess types, the following
wording is typical:
@begin(quotation)
In the case of array types, a check is made that any @ux[constraint
on] the component subtype is the same for the operand array type as
for the target array type. [4.6(13)]
@end(quotation)
The wording here once again shows that to see what a constraint is @dq"on," you
must understand what set of values is being restricted.

Next, let's consider constraints on scalar components of arrays and records.
The wording of 13.2 and 13.4 is @cbon()clear:@cboff() if a bound in a component's range
constraint is not a static expression, a size specification cannot be given
for the enclosing record or array type, nor, in the case of record
components, can a component clause be given.  Although it may sometimes be
inconvenient to provide a static constraint so a representation clause can be
written, the current rules do not make representation clauses so unusable as
to be inconsistent with the intended use of these clauses, i.e., it is not
urgently necessary to correct the rules in order to satisfy the intent of the
design.

Finally, let us consider constraints on components that depend on a
discriminant.  If a value of a constraint is given by a discriminant, the
constraint is not static, since a discriminant is not a constant declared by
a constant declaration (see 4.9(6)).  Hence, no size specification can be
given for type REC, and no component clause can be given for component C of
type REC.

@begin(appendix)
@subheading[!appendix 90-09-07]
@begin(sectionGroup)
@mybar()

!section 13.02    (06) Software Leverage, Inc.  84-10-10              83-00456
!version 1983
!topic Constraints on Access Types
@end(sectionGroup)

LRM 13.2(6) says that a size specification ("for T'Size use...") is "only
allowed if the constraints on T and on its subcomponents (if any) are static".

Our question is, is an index or discriminant constraint associated with a
component of an access type considered to be a constraint on the subcomponent,
or merely of the object (if any) designated by the value of the subcomponent?
For example:

    with Random_Boolean; -- a function with the obvious functionality
    package PK is
      type H(D: Boolean) is
        record
          null;
        end record;
      type T is access H(D => Random_Boolean);
      type R is
        record
          C: T;
        end record;
      for R'Size use 100; -- legal?
    end PK;

The language in LRM 3.8 and 4.8 seems to avoid calling such index and
discriminant constraints constraints on the access value itself, but it isn't
clear that this was intentional.

This question also bears on the rule given in 13.4(7) for component clauses
within record representation clauses.

@begin(sectionGroup)
@mybar()

!section 3.4, 13   Ron Brender   87-08-20
!version 1983
!topic Rep clauses and types derived from formal types
@end(sectionGroup)
!reference AI-00361/09

The compendium of questions raised in AI-00361 seem impressively long,
but  actually  there are a few more to be added.  These all have to do
with representation clauses that apply to types derived from a generic
formal type.  That is, in variations of this schema:

        generic
            type GFT is ...;
        package GP is

            type NT is new GFT;
            for  NT ...;

        end;

...

For SIZE, it is only required that any constraints on the  named  type
or  first  named  subtype  (here,  NT) or any of its subcomponents (if
any), must be static.  In general, this  requirement  seems  trivially
true, yet clearly unacceptable!  To be more specific, consider:

        generic
            type GFT1 is range <>;
            type GFT2 is private;
        package GP is

            type NT1A is new GFT1;
            for  NT1A'SIZE use 16;              -- Legal?

            type NT1B is new GFT1 range 1..10;
            for  NT1B'SIZE use 16;              -- Legal?

            type GFT2 is new GFT2;
            for  GFT2'SIZE use 64;              -- Legal?

        end;

I think all of these ought to be considered illegal, but am not  clear
on  just  what  rationale can lead to this conclusion!  Similar issues
arise wrt SMALL.

@begin(sectionGroup)
@mybar()

!section 13.02 (06)  John Goodenough 89-09-29
!version 1983
!topic Extract from minutes of June 1989 ARG meeting
@end(sectionGroup)

There  was  general  agreement that a representation clause is allowed for an
access type even  if  a  nonstatic  constraint  has  been  specified  in  its
declaration.   Similarly, there was general agreement that a record component
clause could be given for any component of an access type, regardless of  any
constraint  specified in the component's declaration or in the declaration of
the component's type mark.

As for the question concerning a scalar record component  or  type  having  a
nonstatic   constraint,   there   was   no   agreement   concerning   whether
representation clauses should be allowed  under  some  conditions.    It  was
suggested  that  it  was always safe to allow a record component clause for a
component X of type T as long as the specified size was not less than T'SIZE.
Also, the following example was considered:

     type W is new INTEGER range 0..7;
     for W'SIZE use 3;
     subtype SW is W range 0..J;

     type REC is
         record
             NONSTATIC : SW;
             STATIC    :  W;
         end record;

It  seemed  unreasonable  to  some members of the ARG that a record component
clause was allowed for component STATIC but not for component NONSTATIC, even
though component NONSTATIC need not occupy more than three bits.  It was also
argued that the current rule was more drastic than necessary, since  it  does
not even allow an implementation to support a component clause for NONSTATIC,
even though it could perfectly well do so.  However, there was  no  agreement
that component clauses should be allowed for components like NONSTATIC.

The  final  question  discussed  was whether a representation clause could be
given for NREC, a derived record type with a static discriminant  constraint.
Chebat  argued  that  AI-00422  said a record component clause applied to the
record's base type, and in this case,  the  base  type  clearly  contained  a
component  with a nonstatic constraint.  On the other hand, since the size of
any object of type NREC is fixed, there seemed to be little harm in  allowing
a  component  clause  to  be  applied  to  a  component of type NREC.  It was
suggested that this topic be revisited after reviewing AI-00422.

@begin(sectionGroup)
@mybar()

!section 13.02    (06) J-M Chebat 90-09-07                           83-01393
!version 83
!topic This AI is too restrictive
@end(sectionGroup)
!reference AI-00301/05

It seems to me that, concerning components of scalar types, this AI is more
restrictive than the LRM is.

Consider the following example,

    subtype NS_NS is INTEGER range INTEGER(0) .. 255;
    -- NS_NS is a nonstatic subtype and has nonstatic constraints

    subtype NS_S is NS_NS range 0 .. 3;
    -- NS_S is a nonstatic subtype since its type mark isn't static, but has
    -- static constraints according to LRM 4.9 (11)
    --
    --      "A static range is a range whose bounds are static expressions. A
    --      static range constraint is a range constraint whose range is
    --      static"

    type R is
       record
          C : NS_S;
       end record;

    for R use
       record
          C at 0 range 0..1; -- Legal (LAR: Yes / AI: No)

LRM 13.4(7) says,

    "A component clause is only allowed for a component if any constraint on
    this component and any of its subcomponents is static"

then the component clause seems to be legal.

Where the LRM speaks about static constraints, the AI is speaking of
nonstatic subtypes,

    "if a record component has a nonstatic scalar subtype, no component
    clause can be given for the component"

then according to the AI the constraint is illegal.

It has been argued that the current rule is more drastic than necessary. I
think it is not worth to relax. Nevertheless, it would be an error to make it
more drastic again.

An advantage of the wording of the reference manual is that it unifies the case
of composite and scalar type and also that it treats the case of private
types in the same way: two years ago Alsys argued that in an ACVC dispute that
a component clause was not allowed for a private type but the AVO rejected the
dispute on the basis that, anyway, the component had static constraints...

@end(appendix)
